from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import ShuffleSplit, train_test_split
import numpy as np

def train_val_test_split(x, val_size=0.1, test_size=0.1, random_state=None):
    if type(x)==int:
        x = np.zeros(x)
    ss = ShuffleSplit(n_splits=1, test_size=test_size, random_state=random_state)
    train_val_i, test_i = next(ss.split(x))
    train_i, val_i = train_test_split(train_val_i, test_size=val_size/(1-test_size), random_state=random_state)

    # assert len(set(np.concatenate((train_i, val_i, test_i)))) == len(x)
    return train_i, val_i, test_i

class MultiGridSearchCV():
    '''Perform Grid Search over multiple models

    Parameters
    ----------
    estimators: dict
        Name of model. This name has to correspond to model name of param_grid
    param_grids: dict
        Parameter Grid to Search. It must be dict nested in dict(Double dict).
    **kwargs: other parameters from GridSearchCV
        Look GridSearchCV

    examples of **kwargs
    --------------------
    cv: number of k-fold CV
    scoring: scoring method. either can be prebuilt sklearn string or a callable function

    '''
    def __init__(self, estimators, param_grids, **kwargs):
        self.estimators = estimators
        self.param_grids = param_grids
        self.kwargs = kwargs
        self.grid_searches = {estimator_name: GridSearchCV(estimator=estimator, param_grid=self.param_grids[estimator_name], **self.kwargs) for estimator_name, estimator in self.estimators.items()}

    def fit(self, x, y):
        for grid_search in self.grid_searches.values():
            grid_search.fit(x,y)
        self.cv_results = {estimator_name: grid_search.cv_results_ for estimator_name, grid_search in self.grid_searches.items()}
        best_estimator_name_, best_grid_search_ = max(self.grid_searches.items(), key = lambda x: x[1].best_score_) # x[0]: key, x[1]: value
        self.best_score_ = best_grid_search_.best_score_
        self.best_estimator_ = best_grid_search_.best_estimator_
        self.best_params_ = best_grid_search_.best_params_
        self.best_estimator_name_ = best_estimator_name_

    def predict(self, x):
        assert hasattr(self, 'best_estimator_'), 'There is no best_estimator_. Need to call "fit" first.'
        assert hasattr(self.best_estimator_,'predict'), 'Best estimator does not support "predict" method.'
        return self.best_estimator_.predict(x)



class Experiment(object):
    '''
    ------------------
    Data descriptors:
    dirs
        Hold all directory information to be used in the experiment
        in dict() form.
        To see all directories that are related, refer to:
        list(self.dirs.items())
    model_p
        Hold all parameters related to NN model learning in dict() form
    train_p
        Hold all parameters related to training in dict() form
    speaker_list
        Hold all name of speakers
    '''
    def __init__(self, num_speakers = 4, exp_name = None, exp_dir='exp', new = True, model_p = None, train_p = None, lambd = None, debug = False):
        # 0] Random seed
        np.random.seed(0)
        torch.manual_seed(0)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # 1] Hyperparameters setting - Default
        self.model_p = dict(
        vae_lr = 1e-2,
        vae_betas = (0.9,0.999),
        sc_lr = 0.0002,
        sc_betas = (0.5,0.999),
        asr_lr = 0.00001,
        asr_betas = (0.5,0.999),
        ac_lr = 0.00005,
        ac_betas = (0.5,0.999),
        )
        if model_p is not None:
            self.model_p.update(model_p)

        self.train_p = dict(
        n_train_frames = 128,
        batch_size = 32,
        mini_batch_size = 8,
        start_epoch = 1,
        n_epoch = 300,
        model_save_epoch = 3,
        validation_epoch = 3,
        sample_per_path = 10,
        )
        self.train_p['iter_per_ep'] = self.train_p['batch_size'] // self.train_p['mini_batch_size']
        if train_p is not None:
            self.train_p.update(train_p)
        try:
            assert self.train_p['iter_per_ep'] * self.train_p['mini_batch_size'] == self.train_p['batch_size'], 'Specified batch_size "%s" cannot be divided by mini_batch_size "%s"'%(self.train_p['batch_size'], self.train_p['mini_batch_size'])
        except:
            print("Invalid train_p['iter_per_ep'] setting!")
            print('iter_per_ep: %s'%self.train_p['iter_per_ep'])
            print('batch_size: %s'%self.train_p['batch_size'])
            print('mini_batch_size: %s'%self.train_p['mini_batch_size'])
            print('Setting (iter_per_ep) = (batch_size) // (mini_batch_size)')
            self.train_p['iter_per_ep'] = self.train_p['batch_size'] // self.train_p['mini_batch_size']
        self.train_p['epoch'] = self.train_p['start_epoch'] - 1

        self.lambd = dict(
        KLD = 1,
        rec = 20,
        SI = 0,
        LI = 0,
        AC = 0,
        SC = 0,
        C = 0,
        CC = 0,
        )
        if lambd is not None:
            self.lambd.update(lambd)
        self.lambd_total = sum(self.lambd.values())
        if self.lambd['C'] is not 0:
            self.lambd_total -= self.lambd['C']
            self.lambd_total += self.lambd['C'] * (self.lambd['KLD'] + self.lambd['rec'])
        self.lambda_norm = True

        self.preprocess_p = dict(
        sr = 16000,
        frame_period = 5.0,
        num_mcep = 36,
        )
        self.loss_index = ['loss_VAE','loss_KLD','loss_rec','loss_SI','loss_LI','loss_AC','loss_SC','loss_C_KLD', 'loss_C_rec']
        self.performance_measure_index = ['mcd', 'msd_vector', 'gv']
        self.lr_index = ['VAE_lr']
        self.loss_summary = pd.DataFrame(columns = self.loss_index)
        self.validation_summary = pd.DataFrame(columns = self.performance_measure_index)
        self.lr_summary = pd.DataFrame(columns = self.lr_index)

        self.model_kept = []
        self.max_keep=100

        # 2] Initialize environment and variables
        self.create_env(exp_dir = exp_dir, exp_name = exp_name, new = new)
        # self.speaker_list = sorted(os.listdir(self.dirs['train_data']))
        self.speaker_list = ['p225','p226','p227','p228']
        self.num_speakers = len(self.speaker_list)
        assert self.num_speakers == num_speakers, 'Specified "num_speakers" and "num_speakers in train data" does not match'
        self.build_model(params = self.model_p)
        self.p = Printer(filewrite_dir = self.dirs['log'])
        if debug == False:
            sys.stdout = open(self.dirs['log_all'], 'a')
        append(self.dirs['loss_log'], 'epoch '+' '.join(self.loss_index)+'\n')
        append(self.dirs['validation_log'], 'epoch '+' '.join(self.performance_measure_index)+'\n')

        # 3] Hyperparameters for saving model


        # 4] If the experiment is not new, Load most recent model
        if new == False:
            self.model_kept= sorted(os.listdir(self.dirs['model']), key = lambda x: int(x.split('.')[0].split('_')[-1]))
            most_trained_model = self.model_kept[-1]
            epoch_trained = int(most_trained_model.split('_')[-1].split('.')[0])
            self.train_p['start_epoch'] += epoch_trained
            # Update lr_scheduler
            print('Loading model from %s'%most_trained_model)
            self.load_model_all(self.dirs['model'], epoch_trained)

    def create_env(self, exp_dir = 'exp', exp_name = None, new = True):
        '''Create experiment environment
        Store all "static directories" required for experiment in "self.dirs"(dict)

        Store every experiment result in: exp/exp_name/ == exp_dir
        including log, model, test(validation) etc
        '''
        # 0] exp_dir == master directory
        self.dirs = dict()
        # exp_dir = 'exp/'
        model_dir = 'model/'

        # 1] Set up Experiment directory
        if exp_name == None:
            exp_name = time.strftime('%m%d_%H%M%S')
        self.dirs['exp'] = os.path.join(exp_dir, exp_name)
        if new == True:
            assert not os.path.isdir(self.dirs['exp']), 'New experiment, but exp_dir with same name exists'
            os.makedirs(self.dirs['exp'])
        else:
            assert os.path.isdir(self.dirs['exp']), 'Existing experiment, but exp_dir doesn\'t exist'

        # 2] Model parameter directory

    def save_log(self, result, log_dir):
        # 1. Write to log
        log_content = str(self.train_p['epoch'])
        for value in result.mean():
            log_content += ' '+str(value)
        append(log_dir, log_content+'\n')
        # 2. Print result statistics
        self.p.print('Mean\n' + str(result.mean().to_frame().T))
        self.p.print('Std\n' + str(result.std().to_frame().T))

    def save_plot(self, summary, plot_dir):
        for measure in summary.columns:
            fig_save_dir = os.path.join(plot_dir, measure+'.png')
            axes = summary.plot(y = measure, style='o-')
            fig = axes.get_figure()
            fig.savefig(fig_save_dir)
        plt.close('all')

    def performance_measure(self):
        pass



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

class DNN(nn.Module):
    '''
    basic DNN module

    Parameters
    ----------
    n_input: number of input (int)
    n_hidden_list: list of hidden neurons (list of int)
    activation_list: torch.nn activation function instances (nn activation instance or list)

    Example
    -------
    >>> model = DNN(n_input=10, n_hidden_list=[8,6,5], activation_list=[nn.Sigmoid(), nn.ReLU(), nn.Tanh()])
    # n_hidden_list, activation_list corresponds to [h1, h2, output]
    >>> print(model)
    DNN(
      (fc): Sequential(
        (0): Linear(in_features=10, out_features=8, bias=True)
        (1): Sigmoid()
        (2): Linear(in_features=8, out_features=6, bias=True)
        (3): ReLU()
        (4): Linear(in_features=6, out_features=5, bias=True)
        (5): Tanh()
      )
    )-
    '''
    def __init__(self, n_input, n_hidden_list, activation_list):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [activation_list]*len(n_hidden_list)
        assert len(activation_list)==len(n_hidden_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'

        # 1st layer
        layers = [nn.Flatten(1), nn.Linear(n_input, n_hidden_list[0]), activation_list[0]]
        # Hidden layers ~ Output layer
        for i in range(len(n_hidden_list) - 1):
            layers.extend([nn.Linear(n_hidden_list[i], n_hidden_list[i+1]), activation_list[i+1]])

        self.fc = nn.Sequential(*layers)

    def forward(self, x):
        '''x.shape()==(batch_size, feature_dim)'''
        return self.fc(x)

class DNN_Resnet(nn.Module):
    '''
    Resnet module

    Parameters
    ----------
    n_input : int
        number of input
    n_hidden_list : list of int
        list of hidden neurons
    activation : nn activation instance or list
        torch.nn activation function instances
    skip: int
        number of layers to skip. Interval for skipping
        ex) skip=2, then add skip layer every other 2 layers

    Example
    -------
    >>> model = Resnet(n_input=10, n_hidden_list=[8,6,5], activation=[nn.Sigmoid(), nn.ReLU(), nn.tanh()], skip=2)
    # activation corresponds to [h1, h2, output]
    >>> print(model)

    '''
    def __init__(self, n_input, n_hidden_list, activation_list, skip=2):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [activation_list]*len(n_hidden_list)
        assert len(activation_list)==len(n_hidden_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'
        assert len(n_hidden_list) >= skip, 'number of layers must be equal or greater than skip. len(n_hidden_list): %s, skip: %s'%(len(n_hidden_list), skip)
        assert skip >= 2, 'skip needs to be: skip >= 2, given: %s'%(skip)

        # 1st layer
        layers = [nn.Linear(n_input, n_hidden_list[0])]
        # Hidden layers ~ Output layer
        layers.extend([nn.Linear(n_hidden_list[i], n_hidden_list[i+1]) for i in range(len(n_hidden_list)-1)])
        self.fc = nn.ModuleList(layers)
        self.activation_list = nn.ModuleList(activation_list)

        # Skip layers
        self.skip = skip
        n_skip_hidden_list = [n_input] + [n_hidden_list[i] for i in range(skip-1, len(n_hidden_list), skip)]
        skip_layers = ([nn.Linear(n_skip_hidden_list[i], n_skip_hidden_list[i+1], bias=False) for i in range(len(n_skip_hidden_list)-1)]) # Bias already present at original layers
        self.skip_layers = nn.ModuleList(skip_layers) # Need ModuleList, not list, so that the model can recognize this layers
        self.n_skip = len(n_hidden_list) // skip # number of skip layers
        # assert self.n_skip == len(self.skip_layers), 'something\'s wrong'

    def forward(self, x):
        x = x.flatten(1)
        last_x = x
        skip_layers = iter(self.skip_layers)
        for i, (layer, activation) in enumerate(zip(self.fc, self.activation_list)):
            if (i+1) % self.skip == 0:
                skip_layer = next(skip_layers)
                x = activation(layer(x)+skip_layer(last_x))
                last_x = x
            else:
                x = activation(layer(x))
        return x
    '''To do:
    - weight initialization? modify?
    - optimize using nn.Sequential() and itertools in forward. compute layers in group
    '''

class DNN_Densenet(nn.Module):
    def __init__(self, n_input, n_hidden_list, activation_list, skip=2):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [activation_list]*len(n_hidden_list)
        assert len(activation_list)==len(n_hidden_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'
        assert len(n_hidden_list) >= skip, 'number of layers must be equal or greater than skip. len(n_hidden_list): %s, skip: %s'%(len(n_hidden_list), skip)
        assert skip >= 2, 'skip needs to be: skip >= 2, given: %s'%(skip)

        # 1st layer
        layers = [nn.Linear(n_input, n_hidden_list[0])]
        # Hidden layers ~ Output layer
        layers.extend([nn.Linear(n_hidden_list[i], n_hidden_list[i+1]) for i in range(len(n_hidden_list)-1)])
        self.fc = nn.ModuleList(layers)
        self.activation_list = nn.ModuleList(activation_list)

        # Skip layers
        self.skip = skip
        n_skip_hidden_list = [n_input] + [n_hidden_list[i] for i in range(skip-1, len(n_hidden_list), skip)]
        skip_layers = nn.ModuleDict()
        for i in range(1, len(n_skip_hidden_list)):
            for j in range(i):
                skip_layers[str(j)+'_'+str(i)]=nn.Linear(n_skip_hidden_list[j], n_skip_hidden_list[i], bias=False)
        self.skip_layers = skip_layers

        n = len(n_skip_hidden_list)-1
        self.n_skip = int(n*(n+1)/2)
        assert self.n_skip == len(skip_layers), f'something wrong: n_skip({self.n_skip}), skip_layers({len(skip_layers)})'

    def forward(self, x):
        x_history = [x]
        dest = 1 # Destination index of skip layers
        for i, (layer, activation) in enumerate(zip(self.fc, self.activation_list)):
            if (i+1) % self.skip == 0:
                assert dest == len(x_history), f'Length of x_history{len(x_history)} and destination number{dest} must match'
                skip_x = torch.stack([self.skip_layers[str(src)+'_'+str(dest)](x_) for src, x_ in enumerate(x_history)], dim=-1).sum(dim=-1)
                x = activation(layer(x)+skip_x)
                x_history.append(x)
                dest += 1
            else:
                x = activation(layer(x))

        return x

class CNN(nn.Module):
    def __init__(self, in_channel, n_channel_list, activation_list, kernel_size=3):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [activation_list]*len(n_channel_list)
        assert len(activation_list)==len(n_channel_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'

        # 1st layer
        padding = kernel_size // 2
        layers = [nn.Conv2d(in_channels=in_channel, out_channels=n_channel_list[0], kernel_size=kernel_size, stride=1, padding=padding), activation_list[0]]
        # Hidden layers ~ Output layer
        for i in range(len(n_channel_list) - 1):
            layers.extend([nn.Conv2d(in_channels=n_channel_list[i], out_channels=n_channel_list[i+1], kernel_size=kernel_size, stride=1, padding=padding), activation_list[i+1]])
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)

class CNN_Resnet(nn.Module):
    def __init__(self, in_channel, n_channel_list, activation_list, kernel_size=3, skip=2):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [activation_list]*len(n_channel_list)
        assert len(activation_list)==len(n_channel_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'
        assert kernel_size % 2 == 1, f'kernel size must be odd number, Got: ({kernel_size})'

        # 1st layer
        padding = kernel_size // 2
        layers = [nn.Conv2d(in_channels=in_channel, out_channels=n_channel_list[0], kernel_size=kernel_size, stride=1, padding=padding)]
        # Hidden layers ~ Output layer
        layers.extend([nn.Conv2d(in_channels=n_channel_list[i], out_channels=n_channel_list[i+1], kernel_size=kernel_size, stride=1, padding=padding) for i in range(len(n_channel_list)-1)])
        self.layers = nn.ModuleList(layers)
        self.activation_list = nn.ModuleList(activation_list)

        # Skip layers
        self.skip = skip
        n_skip_channel_list = [in_channel]+[n_channel_list[i] for i in range(skip-1, len(n_channel_list), skip)]
        skip_layers = [nn.Conv2d(in_channels=n_skip_channel_list[i], out_channels=n_skip_channel_list[i+1], kernel_size=kernel_size, stride=1, padding=padding, bias=False) for i in range(len(n_skip_channel_list)-1)] # Bias already present at original layers
        self.skip_layers = nn.ModuleList(skip_layers) # Need ModuleList, not list, so that the model can recognize this layers
        self.n_skip = len(n_channel_list) // skip # number of skip layers
        # assert self.n_skip == len(self.skip_layers), 'something\'s wrong'

    def forward(self, x):
        last_x = x
        skip_layers = iter(self.skip_layers)
        for i, (layer, activation) in enumerate(zip(self.layers, self.activation_list)):
            if (i+1) % self.skip == 0:
                skip_layer = next(skip_layers)
                x = activation(layer(x)+skip_layer(last_x))
                last_x = x
            else:
                x = activation(layer(x))
        return x

class CNN_Densenet(nn.Module):
    def __init__(self, in_channel, n_channel_list, activation_list, kernel_size=3, skip=2):
        super().__init__()
        if type(activation_list) is not list:
            activation_list = [activation_list]*len(n_channel_list)
        assert len(activation_list)==len(n_channel_list), 'length of layers and activations must match. If you want no activation, use nn.Identity'
        assert kernel_size % 2 == 1, f'kernel size must be odd number, Got: ({kernel_size})'

        # 1st layer
        padding = kernel_size // 2
        layers = [nn.Conv2d(in_channels=in_channel, out_channels=n_channel_list[0], kernel_size=kernel_size, stride=1, padding=padding)]
        # Hidden layers ~ Output layer
        layers.extend([nn.Conv2d(in_channels=n_channel_list[i], out_channels=n_channel_list[i+1], kernel_size=kernel_size, stride=1, padding=padding) for i in range(len(n_channel_list)-1)])
        self.layers = nn.ModuleList(layers)
        self.activation_list = nn.ModuleList(activation_list)

        # Skip layers
        self.skip = skip
        n_skip_channel_list = [in_channel]+[n_channel_list[i] for i in range(skip-1, len(n_channel_list), skip)]

        skip_layers = nn.ModuleDict()
        for i in range(1, len(n_skip_channel_list)):
            for j in range(i):
                skip_layers[str(j)+'_'+str(i)]=nn.Conv2d(in_channels=n_skip_channel_list[j], out_channels=n_skip_channel_list[i], kernel_size=kernel_size, stride=1, padding=padding, bias=False)
        self.skip_layers = skip_layers

        n = len(n_skip_channel_list)-1
        self.n_skip = int(n*(n+1)/2)
        assert self.n_skip == len(skip_layers), f'something wrong: n_skip({self.n_skip}), skip_layers({len(skip_layers)})'

    def forward(self, x):
        x_history = [x]
        dest = 1 # Destination index of skip layers
        for i, (layer, activation) in enumerate(zip(self.layers, self.activation_list)):
            if (i+1) % self.skip == 0:
                assert dest == len(x_history), f'Length of x_history{len(x_history)} and destination number{dest} must match'
                skip_x = torch.stack([self.skip_layers[str(src)+'_'+str(dest)](x_) for src, x_ in enumerate(x_history)], dim=-1).sum(dim=-1)
                x = activation(layer(x)+skip_x)
                x_history.append(x)
                dest += 1
            else:
                x = activation(layer(x))

        return x

# def residual_computation(sequential, x, connection='sequential'):
#     if connection=='sequential':
#         pass
#     elif connection=='dense':
#         pass
#
#     return

if __name__ == '__main__':
    n_input = 5
    n_hidden_list = [10,20,30,20,10]
    activation_list = [nn.Sigmoid(), nn.ReLU(), nn.LeakyReLU(0.03), nn.Tanh(), nn.Identity()]
    dnn = DNN(n_input, n_hidden_list, activation_list)
    dnn_resnet = DNN_Resnet(n_input, n_hidden_list, activation_list)
    dnn_densenet = DNN_Densenet(n_input, n_hidden_list, activation_list)
    print(dnn)
    print(dnn_resnet)
    print(dnn_densenet)



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

class attention(nn.Module):
    def __init__(self, dim=1):
        super(att_pool, self).__init__()
        self.attention = nn.softmax()


    def forward(self, x):
        return x*self.attention(x)



import torch
import pandas as pd
from torch import Tensor

def to_csv(tensor: torch.Tensor, path: str):
    assert type(tensor)==torch.Tensor
    assert len(tensor.shape) <= 2, f'Must pass 2-d input. shape={tensor.shape}'
    pd.DataFrame(tensor.cpu().numpy()).to_csv(path, index=False)



import os
import pickle
import time
import subprocess
from pathlib import Path as P
'''import tools.tools as tools'''

def save_pickle(obj: str, path: str = None):
    '''Save object as Pickle file to designated path.
    If path is not given, default to "YearMonthDay_HourMinuteSecond.p" '''
    if path == None:
        path = time.strftime('%y%m%d_%H%M%S.p')
    with open(path, 'wb') as f:
        pickle.dump(obj, f)

def load_pickle(path: str):
    '''Load Pickle file from designated path'''
    with open(path, 'rb') as f:
        return pickle.load(f)

def write(content, path, encoding = None):
    with open(path, 'w', encoding = encoding) as f:
        f.write(content)

def append(content, path, encoding = None):
    with open(path, 'a', encoding = encoding) as f:
        f.write(content)

def read(path, encoding = None):
    with open(path, 'r', encoding = encoding) as f:
        text = f.read()
    return text

def readline(path, encoding = None):
    '''Create generator object which iterates over the file'''
    with open(path, 'r', encoding = encoding) as f:
        for line in f:
            yield line

def readlines(path, encoding = None):
    with open(path, 'r', encoding = encoding) as f:
        text = f.readlines()
    return text

def cmd(command: str):
    '''Run shell command'''
    pipe = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding='utf-8')
    return pipe.stdout

def prettify_dict(dictionary, indent=0):
    return '\n'.join([' '*indent + str(k) +': '+str(v) if type(v)!=dict else str(k)+':\n'+prettify_dict(v, indent=indent+2) for k, v in dictionary.items()])

# Used in Argparse
def str2bool(x):
    true_list = ['t', 'true', 'y', 'yes', '1']
    false_list = ['f', 'false', 'n', 'no', '0']
    if x.lower() in true_list:
        return True
    elif x.lower() in false_list:
        return False
    else:
        raise Exception('input has to be in one of two forms:\nTrue: %s\nFalse: %s'%(true_list, false_list))

def is_iterable(x):
    return hasattr(x, '__iter__')

def now(format: str ='-'):
    if format=='-':
        return time.strftime('%Y-%m-%d_%H-%M-%S')
    elif format=='_':
        return time.strftime('%y%m%d_%H%M%S')
    else:
        raise Exception("format has to be one of ['-', '_']")

import yaml
def print_dict(d, **kwargs):
    '''prettify long dictionary
    Example
    -------
    >>> d = {'a':1, 'b':2, 'c':3}
    >>> print_dict(d)

    '''

    return yaml.dumps(d, **kwargs)

class Filename():
    '''(Will be deprecated since Path() can be used)
    Class to handle Filename with suffix.
    To call filename with suffix, call the instance.

    Example
    -------
    >>> file = Filename('model', '.pt')
    >>> file
    (Filename, name: "model", suffix: ".pt")
    >>> str(file)
    "model"
    >>> file()
    "model.pt"
    >>> (file+'1')
    "model1"
    >>> (file+'1')()
    "model1.pt"
    '''
    def __init__(self,obj=None, suffix=''):
        self.name = str(obj)
        self.suffix = suffix

    def __call__(self):
        '''Add suffix and return str'''
        return self.name+self.suffix

    def __radd__(self, other):
        return Filename(other+self.name,suffix=self.suffix)

    def __add__(self, other):
        return Filename(self.name+other,suffix=self.suffix)

    def __mul__(self, other):
        return Filename(self.name*other,suffix=self.suffix)

    def __repr__(self):
        return '(Filename, name: "%s", suffix: "%s")'%(self.name, self.suffix)

    def __str__(self):
        return self.name

class Path(str):
    '''
    Joins paths by . syntax
    Use pathlib.Path internally

    Parameters
    ----------
    path: str (default: '.')
        Notes the default path. Leave for default blank value which means the current working directory.

    Example
    -------
    >>> path = Path('C:/exp')
    >>> path
    path: C:/exp

    >>> path.DATA = 'CelebA'
    >>> path
    path: C:/exp
    DATA: C:/exp/CelebA

    >>> path.PROCESSED = 'processed'
    >>> path.PROCESSED.M1 = 'method1'
    >>> path.PROCESSED.M2 = 'method2'
    >>> path
    path: C:/exp
    DATA: C:/exp/CelebA
    PROCESSED: C:/exp/processed

    >>> path.PROCESSED
    M1: C:/exp/processed/method1
    M2: C:/exp/processed/method2
    -------

    '''
    def __init__(self, path='.'):
        self.path=path

    def __repr__(self):
        return self.path


    def __call__(self):
        '''Print out current path, and children'''
        for name, directory in self.__dict__.items():
            if name != 'path':
                print(name+': '+str(directory))
                if type(directory) == Path:
                    directory()
        # print('\n'.join([key+': '+str(value) for key, value in self.__dict__.items()]))

    def __str__(self):
        return self.path

    def __setattr__(self, key, value):
        # super(Path, self).__setattr__(key, self / value) # self.joinpath(value)
        if hasattr(self, 'path'):
            super(Path, self).__setattr__(key, Path(os.path.join(self.path, value)))
        else:
            super(Path, self).__setattr__(key, value)

    def join(self, *args):
        return os.path.join(self.path, *args)

    def makedirs(self, exist_ok=True):
        '''Make directories of all children paths
        Be sure to define all folders first, makedirs(), and then define files in Path(),
        since defining files before makedirs() will lead to creating directories with names of files.
        It is possible to ignore paths with "." as all files do, but there are hidden directories that
        start with "." which makes things complicated. Thus, defining folders -> makedirs() -> define files
        is recommended.'''
        for directory in self.__dict__.values():
            if directory != '':
                os.makedirs(str(directory), exist_ok=exist_ok)
                if type(directory) == Path:
                    directory.makedirs(exist_ok=exist_ok)

class tdict(dict):
    '''
    Dictionary which can get items via attribute notation (class.attribute)

    Parameters
    ----------
    Identical to dict()

    Example
    -------

    '''
    def __init__(self, *args, **kwargs):
        super(tdict, self).__init__(*args, **kwargs)

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __delattr__(self, key):
        self.__delitem__(key)

class Printer():
    def __init__(self, filewrite_dir = None):
        self.content = ''
        self.filewrite_dir = filewrite_dir

    def add(self, text):
        self.content += text

    def print(self, *args, end='\n', flush=False):
        self.add(' '.join([str(arg) for arg in args]))
        print(self.content, end=end, flush=flush)
        if self.filewrite_dir != None:
            append(self.content + end, self.filewrite_dir)
        self.content=''

    def reset(self):
        self.content = ''

class Timer():
    '''Timer to measure elapsed time

    Parameters
    ----------
    print: bool (default: False)
        if True, then prints self.elapsed_time whenever stop() is called.

    return_f: bool (default: True)
        if True, then returns self.elapsed_time whenever stop() is called.

    auto_reset: bool (default: True)
        if True, then resets whenever start() is called.

    Methods
    -------
    start

    stop

    reset
        sets self.elapsed_time = 0
    '''
    def __init__(self, print = False, return_f = True, auto_reset = True):
        self.print = print
        self.return_f = return_f
        self.auto_reset = auto_reset
        self.elapsed_time = 0
        self.running = False

    def __enter__(self):
        self.start()

    def __exit__(self):
        t = self.stop()
        print(t)

    def start(self):
        if self.auto_reset:
            # self.reset()
            self.elapsed_time = 0
        self.running = True
        self.start_time = time.time()

    def stop(self):
        if self.running:
            self.end_time = time.time()
            self.elapsed_time += self.end_time - self.start_time
            self.running = False
        if self.print:
            print(self.elapsed_time)
        if self.return_f:
            return self.elapsed_time

    def reset(self):
        self.elapsed_time = 0

class AverageMeter(object):
    """Computes and stores the average and current value
    Variables
    ---------
    self.val
    self.avg
    self.sum
    self.count
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def step(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count




from . import federated_learning, model
from ._pandas import *

def device_of(nn_module):
    return next(nn_module.parameters()).device

def device_of_optimizer(optimizer):
    return optimizer.param_groups[0]['params'][0].device

def to_onehot(x, max_dim):
    '''
    x: list of int

    max_dim: maximum dimension
    '''
    onehot = torch.zeros(len(x),max_dim)
    onehot[range(len(x)), x] = 1
    return onehot



import torch
import torch.nn as nn
from sklearn.base import BaseEstimator
from model import DNN
import sys
import time
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class estimator(BaseEstimator):
    def __init__(self, model=DNN(10, [5,3,1], nn.ReLU()), val_dataset='', n_epoch=10, lr=0.001, batch_size=32, device=device, log_dir=None):
        # All arguments need default values.
        # All arguments must be fed into self.args = args
        self.model = model
        self.val_dataset = val_dataset
        self.n_epoch = n_epoch
        self.lr = lr
        self.batch_size = batch_size
        self.device = device
        self.log_dir = log_dir

    def fit(self, x, y):
        if self.log_dir is None:
            log_dir = open(path.EXP.join(time.strftime('%Y-%m-%d_%H-%M-%S')+'.txt'))
        default_stdout = sys.stdout
        sys.stdout = log_dir


        # All parameters created here should be fed into self.args_ = args


    def predict(self, x):
        # return predicted value
        return y



import os
import numpy as np
from .tools import load_pickle

def sort_load(data_dir, load_func = None):
    '''load all data in the specified directory, in a sorted way
    if load_func == None, defaults to pickle'''
    if load_func == None:
        load_func = load_pickle

    file_list = sorted(os.listdir(data_dir))
    loaded = list()

    for file in file_list:
        data_dir_1 = os.path.join(data_dir, file)
        data = load_func(data_dir_1)
        loaded.append(data)

    return loaded

def sample_train_data(dataset_A, dataset_B,ppgset_A,ppgset_B, n_frames=128):

    num_samples = min(len(dataset_A), len(dataset_B))
    train_data_A_idx = np.arange(len(dataset_A))
    train_data_B_idx = np.arange(len(dataset_B))
    np.random.shuffle(train_data_A_idx)
    np.random.shuffle(train_data_B_idx)
    train_data_A_idx_subset = train_data_A_idx[:num_samples]
    train_data_B_idx_subset = train_data_B_idx[:num_samples]

    train_data_ppg_A = list()
    train_data_ppg_B = list()
    train_data_A = list()
    train_data_B = list()

    for idx_A, idx_B in zip(train_data_A_idx_subset, train_data_B_idx_subset):
        data_A = dataset_A[idx_A]
        data_ppg_A = ppgset_A[idx_A]
        frames_A_total = data_A.shape[1]
        frames_A_ppg_total = data_ppg_A.shape[1]
        #print(frames_A_total)
        #print(frames_A_ppg_total)
        #print(min([frames_A_total,frames_A_ppg_total]))
        assert min([frames_A_total,frames_A_ppg_total]) >= n_frames
        start_A = np.random.randint(min([frames_A_total,frames_A_ppg_total]) - n_frames + 1)
        end_A = start_A + n_frames
        train_data_A.append(data_A[:, start_A:end_A])
        train_data_ppg_A.append(data_ppg_A[:, start_A:end_A])

        data_B = dataset_B[idx_B]
        data_ppg_B = ppgset_B[idx_B]
        frames_B_total = data_ppg_B.shape[1]
        frames_B_ppg_total = data_ppg_B.shape[1]
        #print(min([frames_B_total,frames_B_ppg_total]))
        assert min([frames_B_total,frames_B_ppg_total]) >= n_frames
        start_B = np.random.randint(min([frames_B_total,frames_B_ppg_total]) - n_frames + 1)
        end_B = start_B + n_frames
        train_data_B.append(data_B[:, start_B:end_B])
        train_data_ppg_B.append(data_ppg_B[:, start_B:end_B])
        #print(np.shape(data_B))#data_B
    #print(len(train_data_A))
    #print(len(train_data_B))
    train_data_ppg_A = np.array(train_data_ppg_A)
    train_data_ppg_B = np.array(train_data_ppg_B)
    train_data_A = np.array(train_data_A)
    train_data_B = np.array(train_data_B)

    #train_data_A = np.expand_dims(train_data_A, axis=-1)
    #train_data_B = np.expand_dims(train_data_B, axis=-1)
    return train_data_A, train_data_B,train_data_ppg_A,train_data_ppg_B

class Tree():
    '''Not implemented yet'''
    def __init__(self, data = None, parent = None):
        self.data = data

        self.parent = parent
        self.children = list()
        self.depth = 0

    def __call__(self):
        return self.data

    def __getitem__(self, key):
        pass

    def __setitem__(self, key):
        pass
    def __len__(self):
        pass
